# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt Build context module
~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


import codecs
from collections import defaultdict, deque
from concurrent.futures import ThreadPoolExecutor
from functools import reduce
import json
import os
from pathlib import PurePath
import platform
from subprocess import PIPE
import sys
import threading
from time import sleep, time

import networkx as nx
from colorama import Fore, Style
from ostrich.utils.proc import run, CalledProcessError
from ostrich.utils.text import get_safe_path

from yabt.cli import call_user_func
from .caching import (get_prebuilt_targets, load_target_from_cache,
                      save_target_in_cache, save_test_in_cache)
from .config import Config
from .docker import format_qualified_image_name
from .extend import Plugin
from .graph import get_ancestors, get_descendants, topological_sort
from .logging import make_logger
from .target_extraction import extractor
from .target_utils import split_build_module, Target
from .utils import fatal, fatal_noexc


logger = make_logger(__name__)


class BuildContext:
    """Build Context class.

    The framework is designed to operate with a single instance of the
    BuildContext alive, but this is not enforced using a singleton pattern,
    because there's no particular reason to do so...

    While not yet implemented, it is designed to supported concurrent parsing
    of multiple build files using the same build context.
    The design principle to enable this:
    - The "driver" (e.g. main thread in main process) is responsible for
      dispatching build file parsers with copies or references of the build
      context (e.g. on other threads / processes / machines?).
    - Parsers, with their own build context, populate partial target maps.
    - The driver collects the partial results, and is responsible to merge
      them consistently and safely.
    - This will worth it, of course, only if serializing the state to a parser
      and serializing the partial result and merging it back is MUCH cheaper
      than actually doing the parsing, because if it's not, it would be better
      to do it all in the driver (much simpler...) - which is the reason this
      is not yet implemented.
    """

    def __init__(self, conf: Config):
        self.conf = conf
        # A *thread-safe* targets map
        self.targets = {}
        self.failed_nodes = {}
        self.skipped_nodes = []
        # A *thread-safe* map from build module to set of target names
        # that were extracted from that build module
        self.targets_by_module = defaultdict(set)
        # A *thread-safe* set of processed build-files
        self.processed_build_files = set()
        # Target graph is *not necessarily thread-safe*!
        self.target_graph = None
        # # A *thread-safe* map from BuildEnv name to qualified Docker image
        # #  name for that BuildEnv
        # self.buildenv_images = {}
        # A dictionary for collecting metadata on build artifacts
        self.artifacts_metadata = {}
        self.context_lock = threading.Lock()
        self.global_cache = call_user_func(conf.settings, 'get_global_cache')
        self.global_cache_failures = 0

    def get_workspace(self, *parts) -> str:
        """Return a path to a private workspace dir.
           Create sub-tree of dirs using strings from `parts` inside workspace,
           and return full path to innermost directory.

        Upon returning successfully, the directory will exist (potentially
        changed to a safe FS name), even if it didn't exist before, including
        any intermediate parent directories.
        """
        workspace_dir = os.path.join(self.conf.get_workspace_path(),
                                     *(get_safe_path(part) for part in parts))
        if not os.path.isdir(workspace_dir):
            # exist_ok=True in case of concurrent creation of the same dir
            os.makedirs(workspace_dir, exist_ok=True)
        return workspace_dir

    def get_bin_dir(self, build_module: str) -> str:
        """Return a path to the binaries dir for a build module dir.
           Create sub-tree of missing dirs as needed, and return full path
           to innermost directory.
        """
        bin_dir = os.path.join(self.conf.get_bin_path(), build_module)
        if not os.path.isdir(bin_dir):
            # exist_ok=True in case of concurrent creation of the same dir
            os.makedirs(bin_dir, exist_ok=True)
        return bin_dir

    def walk_target_deps_topological_order(self, target: Target):
        """Generate all dependencies of `target` by topological sort order."""
        all_deps = get_descendants(self.target_graph, target.name)
        for dep_name in topological_sort(self.target_graph):
            if dep_name in all_deps:
                yield self.targets[dep_name]

    def generate_direct_deps(self, target: Target):
        """Generate only direct dependencies of `target`."""
        yield from (self.targets[dep_name] for dep_name in sorted(target.deps))

    def generate_dep_names(self, target: Target):
        """Generate names of all dependencies (descendants) of `target`."""
        yield from sorted(get_descendants(self.target_graph, target.name))

    def generate_all_deps(self, target: Target):
        """Generate all dependencies of `target` (the target nodes)."""
        yield from (self.targets[dep_name]
                    for dep_name in self.generate_dep_names(target))

    def register_target(self, target: Target):
        """Register a `target` instance in this build context.

        A registered target is saved in the `targets` map and in the
        `targets_by_module` map, but is not added to the target graph until
        target extraction is completed (thread safety considerations).
        """
        if target.name in self.targets:
            first = self.targets[target.name]
            raise NameError(
                'Target with name "{0.name}" ({0.builder_name} from module '
                '"{1}") already exists - defined first as '
                '{2.builder_name} in module "{3}"'.format(
                    target, split_build_module(target.name),
                    first, split_build_module(first.name)))
        self.targets[target.name] = target
        self.targets_by_module[split_build_module(target.name)].add(
            target.name)

    def remove_target(self, target_name: str):
        """Remove (unregister) a `target` from this build context.

        Removes the target instance with the given name, if it exists,
        from both the `targets` map and the `targets_by_module` map.

        Doesn't do anything if no target with that name is found.

        Doesn't touch the target graph, if it exists.
        """
        if target_name in self.targets:
            del self.targets[target_name]
        build_module = split_build_module(target_name)
        if build_module in self.targets_by_module:
            self.targets_by_module[build_module].remove(target_name)

    def get_target_extraction_context(self, build_file_path: str) -> dict:
        """Return a build file parser target extraction context.

        The target extraction context is a build-file-specific mapping from
        builder-name to target extraction function,
        for every registered builder.
        """
        extraction_context = {}
        for name, builder in Plugin.builders.items():
            extraction_context[name] = extractor(name, builder,
                                                 build_file_path, self)
        return extraction_context

    # def register_buildenv_image(self, name: str, docker_image: str):
    #     """Register a named BuildEnv Docker image in this build context."""
    #     self.buildenv_images[name] = docker_image

    def get_buildenv_graph(self):
        """Return a graph induced by buildenv nodes"""
        # This implementation first obtains all subsets of nodes that all
        # buildenvs depend on, and then builds a subgraph induced by the union
        # of these subsets. This can be very non-optimal.
        # TODO(itamar): Reimplement efficient algo, or redesign buildenvs
        buildenvs = set(target.buildenv for target in self.targets.values()
                        if target.buildenv)
        return nx.DiGraph(self.target_graph.subgraph(reduce(
            lambda x, y: x | set(y),
            (get_descendants(self.target_graph, buildenv)
             for buildenv in buildenvs), buildenvs)))

    def ready_nodes_iter(self, graph_copy):
        """Generate ready targets from the graph `graph_copy`.

        The input graph is mutated by this method, so it has to be a mutable
        copy of the graph (e.g. not original copy, or read-only view).

        Caller **must** call `done()` after processing every generated
        target, so additional ready targets can be added to the queue.

        The invariant: a target may be yielded from this generator only
        after all its descendant targets were notified "done".
        """

        def is_ready(target_name):
            """Return True if the node `target_name` is "ready" in the graph
               `graph_copy`.

            "Ready" means that the graph doesn't contain any more nodes that
            `target_name` depends on (e.g. it has no successors).
            """
            try:
                next(graph_copy.successors(target_name))
            except StopIteration:
                return True
            return False

        ready_nodes = deque(sorted(
            target_name for target_name in graph_copy.nodes
            if is_ready(target_name)))
        produced_event = threading.Event()
        failed_event = threading.Event()

        def make_done_callback(target: Target):
            """Return a callable "done" notifier to
               report a target as processed."""

            def done_notifier():
                """Mark target as done, adding new ready nodes to queue"""
                if graph_copy.has_node(target.name):
                    affected_nodes = list(sorted(
                        graph_copy.predecessors(target.name)))
                    graph_copy.remove_node(target.name)
                    ready_nodes.extend(
                        target_name for target_name in affected_nodes
                        if is_ready(target_name))
                    produced_event.set()

            return done_notifier

        def make_retry_callback(target: Target):
            """Return a callable "retry" notifier to
               report a target as in need of retry.
               Currently for tests we rebuild the target
               when it's not necessary."""

            def retry_notifier():
                """Mark target as retry, re-entering node to end of queue"""
                if graph_copy.has_node(target.name):
                    ready_nodes.append(target.name)
                    produced_event.set()

            return retry_notifier

        def make_fail_callback(target: Target):
            """Return a callable "fail" notifier to
               report a target as failed after all retries."""

            def fail_notifier(ex):
                """Mark target as failed, taking it and ancestors
                   out of the queue"""
                # TODO(Dana) separate "failed to build target" errors from
                # "failed to run" errors.
                # see: https://github.com/resonai/ybt/issues/124
                try:
                    if isinstance(ex, CalledProcessError) and ex.stdout:
                        # TODO(Dana) When ex.stdout has non ascii chars the
                        # call to sys.stdout.write crashes in the inner
                        # function, when colorama/ansitowin32.py assumes that
                        # the text is ascii encoded.
                        sys.stdout.write(ex.stdout.decode('utf-8'))
                        sys.stderr.write(ex.stderr.decode('utf-8'))
                finally:
                    if graph_copy.has_node(target.name):
                        self.failed_nodes[target.name] = ex
                        # remove all ancestors (nodes that depend on this one)
                        affected_nodes = get_ancestors(graph_copy, target.name)
                        graph_copy.remove_node(target.name)
                        for affected_node in affected_nodes:
                            if graph_copy.has_node(affected_node):
                                if affected_node not in self.skipped_nodes:
                                    self.skipped_nodes.append(affected_node)
                                graph_copy.remove_node(affected_node)
                        if self.conf.continue_after_fail:
                            logger.info('Failed target: {} due to error: {}',
                                        target.name, ex)
                            produced_event.set()
                        else:
                            failed_event.set()
                            fatal('`{}\': {}', target.name, ex)

            return fail_notifier

        while True:
            while len(ready_nodes) == 0:
                if graph_copy.order() == 0:
                    return
                if failed_event.is_set():
                    return
                produced_event.wait(0.5)
            produced_event.clear()
            next_node = ready_nodes.popleft()
            node = self.targets[next_node]
            node.done = make_done_callback(node)
            # TODO(bergden) retry assumes no need to update predecessors:
            # This means we don't support retries for targets that are
            # prerequisites of other targets (builds, installs)
            node.retry = make_retry_callback(node)
            node.fail = make_fail_callback(node)
            yield node

    def target_iter(self):
        """Generate ready targets from entire target graph.

        Caller **must** call `done()` after processing every generated target,
        so additional ready targets can be added to the queue.

        The invariant: a target may be yielded from this generator only after
        all its descendant targets were notified "done".
        """
        yield from self.ready_nodes_iter(self.target_graph.copy())

    def buildenv_iter(self):
        """Generate ready targets from subgraph of buildenvs.

        Caller **must** call `done()` after processing every generated target,
        so additional ready targets can be added to the queue.

        The invariant: a target may be yielded from this generator only after
        all its descendant targets were notified "done".
        """
        yield from self.ready_nodes_iter(self.get_buildenv_graph())

    def run_in_buildenv(
            self, buildenv_target_name: str, cmd: list, cmd_env: dict=None,
            work_dir: str=None, auto_uid: bool=True, run_params: list=None,
            **kwargs):
        """Run a command in a named BuildEnv Docker image.

        :param buildenv_target_name: A named Docker image target in which the
                                     command should be run.
        :param cmd: The command to run, as you'd pass to subprocess.run()
        :param cmd_env: A dictionary of environment variables for the command.
        :param work_dir: A different work dir to run in.
                         Either absolute path, or relative to project root.
        :param auto_uid: Whether to run as the active uid:gid, or as root.
        :param run_params: Params to pass to the docker run command.
        :param kwargs: Extra keyword arguments that are passed to the
                        subprocess.run() call that runs the BuildEnv container
                        (for, e.g. timeout arg, stdout/err redirection, etc.)

        :raises KeyError: If named BuildEnv is not a registered BuildEnv image
        """
        buildenv_target = self.targets[buildenv_target_name]
        # TODO(itamar): Assert that buildenv_target is up to date
        redirection = any(
            stream_key in kwargs
            for stream_key in ('stdin', 'stdout', 'stderr', 'input'))
        docker_run = ['docker', 'run']
        # if not self.conf.non_interactive:
        #     docker_run.append('-i')
        if not redirection:
            docker_run.append('-t')
        project_vol = (self.conf.docker_volume if self.conf.docker_volume else
                       self.conf.project_root)
        container_work_dir = PurePath('/project')
        if work_dir:
            container_work_dir /= work_dir

        docker_run.extend([
            '--rm',
            '-v', project_vol + ':/project',
            # TODO: windows containers?
            '-w', container_work_dir.as_posix(),
        ])
        if cmd_env:
            for key, value in cmd_env.items():
                # TODO(itamar): escaping
                docker_run.extend(['-e', '{}={}'.format(key, value)])
        if platform.system() == 'Linux' and auto_uid:
            # Fix permissions for bind-mounted project dir
            # The fix is not needed when using Docker For Mac / Windows,
            # because it is somehow taken care of by the sharing mechanics
            docker_run.extend([
                '-u', '{}:{}'.format(os.getuid(), os.getgid()),
                '-v', '/etc/shadow:/etc/shadow:ro',
                '-v', '/etc/group:/etc/group:ro',
                '-v', '/etc/passwd:/etc/passwd:ro',
                '-v', '/etc/sudoers:/etc/sudoers:ro',
            ])
        if run_params:
            docker_run.extend(run_params)
        docker_run.append(format_qualified_image_name(buildenv_target))
        docker_run.extend(cmd)
        logger.info('Running command in build env "{}" using command {}',
                    buildenv_target_name, ' '.join(
                        format_for_cli(part) for part in docker_run))
        # TODO: Consider changing the PIPEs to temp files.
        if 'stderr' not in kwargs:
            kwargs['stderr'] = PIPE
        if 'stdout' not in kwargs:
            kwargs['stdout'] = PIPE
        result = run(docker_run, check=True, **kwargs)

        # TODO(Dana): Understand what is the right enconding and remove the
        # try except
        if kwargs['stdout'] is PIPE:
            try:
                sys.stdout.write(result.stdout.decode('utf-8'))
            except UnicodeEncodeError as e:
                sys.stderr.write('tried writing the stdout of {},\n but it '
                                 'has a problematic character:\n {}\n'
                                 'partial hex dump of stdout:\n{}\n'
                                 .format(docker_run, str(e),
                                         codecs.encode(result.stdout, 'hex')
                                         .decode('utf8')[:1000]))
        if kwargs['stderr'] is PIPE:
            try:
                sys.stderr.write(result.stderr.decode('utf-8'))
            except UnicodeEncodeError as e:
                sys.stderr.write('tried writing the stderr of {},\n but it '
                                 'has a problematic character:\n {}\n'
                                 'partial hex dump of stderr:\n{}\n'
                                 .format(docker_run, str(e),
                                         codecs.encode(result.stderr, 'hex')
                                         .decode('utf8')[:1000]))
        return result

    def build_target(self, target: Target):
        """Invoke the builder function for a target."""
        builder = Plugin.builders[target.builder_name]
        if builder.func:
            logger.debug('About to invoke the {} builder function for {}',
                         target.builder_name, target.name)
            builder.func(self, target)
        else:
            logger.debug('Skipping {} builder function for target {} (no '
                         'function registered)', target.builder_name, target)

    def test_target(self, target: Target):
        """Invoke the tester function for a target."""
        builder = Plugin.builders[target.builder_name]
        if builder.test_func:
            logger.debug('About to invoke the {} tester function for {}',
                         target.builder_name, target.name)
            builder.test_func(self, target)
        else:
            logger.debug('Skipping {} tester function for target {} (no '
                         'function registered)', target.builder_name, target)

    def register_target_artifact_metadata(self, target: str, metadata: dict):
        """Register the artifact metadata dictionary for a built target."""
        with self.context_lock:
            self.artifacts_metadata[target.name] = metadata

    def write_artifacts_metadata(self):
        """Write out a JSON file with all built targets artifact metadata,
           if such output file is specified."""
        if self.conf.artifacts_metadata_file:
            logger.info('Writing artifacts metadata to file "%s"',
                        self.conf.artifacts_metadata_file)
            with open(self.conf.artifacts_metadata_file, 'w') as fp:
                json.dump(self.artifacts_metadata, fp)

    def can_use_cache(self, target: Target) -> bool:
        """Return True if should attempt to load `target` from cache.
           Return False if `target` has to be built, regardless of its cache
           status (because cache is disabled, or dependencies are dirty).
        """
        # if caching is disabled for this execution, then all targets are dirty
        if self.conf.no_build_cache:
            return False
        # if the target's `cachable` prop is falsy, then it is dirty
        if not target.props.cachable:
            return False
        # if any dependency of the target is dirty, then the target is dirty
        if any(self.targets[dep].is_dirty for dep in target.deps):
            logger.info('Cannot use cache of target: {} because it has dirty '
                        'dependencies: {}', target.name,
                        [dep for dep in target.deps
                         if self.targets[dep].is_dirty])
            return False
        # if the target has a dirty buildenv then it's also dirty
        if target.buildenv and self.targets[target.buildenv].is_dirty:
            logger.info('Cannot use cache of target: {} because its buildenv '
                        'is dirty: {}', target.name, target.buildenv)
            return False
        return True

    def build_graph(self, run_tests: bool=False):
        built_targets = set(get_prebuilt_targets(self))

        def build(target: Target):
            """Build `target` if it wasn't built already, and mark it built."""
            # avoid rebuilding built target
            if target.name in built_targets:
                target.done()
                return
            try:
                # call pre-build hook, if one exists
                if hasattr(target, 'pre_build_hook'):
                    logger.debug('Calling pre-build hook for target {}',
                                 target.name)
                    pre_build_hook = getattr(target, 'pre_build_hook')
                    pre_build_hook(self, target)
                # note: running compute hash here so the hash is not affected
                # by the build func itself (and may change based on whether
                # the build was cached or not), but not earlier than here,
                # so all previous nodes have already been built, including
                # artifacts that this target may rely on in hacky ways, such
                # as following ExtCommand or Grunt builder with a FileGroup
                # builder that collects generated files that are not handled
                # as "real artifacts"...
                # TODO: fix this...
                target.compute_hash(self)
                logger.info('Target {} hash: {}',
                            target.name, target.hash(self))
                # check if cache can be used to skip building target
                build_cached, test_cached = False, False
                if self.can_use_cache(target):
                    build_cached, test_cached = load_target_from_cache(
                        target, self)

                target_built = False
                if build_cached:
                    logger.info('Target {} loaded from cache - skipping build',
                                target.name)
                else:
                    logger.info('Building target {}', target.name)
                    target.is_dirty = True
                    # test can't be cached if running build
                    test_cached = False
                    build_start = time()
                    self.build_target(target)
                    target.summary['build_time'] = time() - build_start
                    target.info['test_time'] = None
                    target.tested = {}
                    logger.info('Build of target {} completed in {} sec',
                                target.name, target.summary['build_time'])
                    target_built = True
                # write to cache only if build was executed
                if target_built:
                    save_target_in_cache(target, self)

                # TODO: collect stats and print report at the end
                target_tested = False
                if run_tests and 'testable' in target.tags:
                    if self.conf.no_test_cache or not test_cached:
                        logger.info('Testing target {}', target.name)
                        test_start = time()
                        self.test_target(target)
                        target.info['test_time'] = time() - test_start
                        target.tested[target.test_hash(self)] = target.info[
                          'test_time']
                        logger.info(
                            'Test of target {} completed in {} sec '
                            'with {} fails',
                            target.name, target.info['test_time'],
                            target.info['fail_count'])
                        target_tested = True
                    else:
                        logger.info(
                            'Target {} test cached - skipping test run',
                            target.name)
                built_targets.add(target.name)
                target.done()

                # write to cache only if test was executed
                if target_tested:
                    save_test_in_cache(target, self)
            except Exception as ex:
                target.info['fail_count'] += 1
                default_attempts = 1
                if 'testable' in target.tags:
                    default_attempts = self.conf.test_attempts
                attempts = max(target.props.attempts, default_attempts)
                if attempts > target.info['fail_count']:
                    target.retry()
                else:
                    target.fail(ex)

        def build_in_pool(seq):
            jobs = self.conf.jobs
            # don't use thread pool in case of single worker
            if jobs > 1:
                with ThreadPoolExecutor(max_workers=jobs) as executor:
                    list(executor.map(build, seq))
            else:
                for target in seq:
                    build(target)

        logger.info('Marked {} targets as "pre-built" in cached base images',
                    len(built_targets))

        logger.info('Building {} flavor using {} workers -- BuildEnv PrePass',
                    self.conf.flavor, self.conf.jobs)
        # pre-pass: build detected buildenv targets and their dependencies
        build_in_pool(self.buildenv_iter())
        logger.info('Building {} flavor using {} workers -- Main Pass',
                    self.conf.flavor, self.conf.jobs)
        # main pass: build rest of the graph
        build_in_pool(self.target_iter())
        if self.failed_nodes:
            print(Fore.RED +
                  '\n=============================',
                  '\n   Finished with failures.',
                  '\n=============================' +
                  Style.RESET_ALL)
            for target_name, ex in self.failed_nodes.items():
                if isinstance(ex, CalledProcessError) and ex.stdout:
                    print('\n\nTarget', target_name,
                          'failed executing command:\n\n')
                    print(' '.join(ex.cmd[0]))
                    print('\n')
                    if ex.stdout:
                        print('\n=============================',
                              '\nstdout output for the target:',
                              '\n=============================\n')
                        print(ex.stdout.decode('utf-8'))
                    if ex.stderr:
                        print('\n=============================',
                              '\nstderr output for the target:',
                              '\n=============================')
                        print(ex.stderr.decode('utf-8'))
                else:
                    print('\n\nTarget', target_name, 'failed with error:',
                          str(ex))
            fatal_noexc('Finished building target graph with fails: \n{}\n'
                        'which caused the following to skip: \n{}',
                        list(self.failed_nodes), self.skipped_nodes)
        else:
            logger.info('Finished building target graph successfully')


def format_for_cli(part):
    return '"{}"'.format(part) if ' ' in part else part



# -*- coding: utf-8 -*-

# Copyright 2018 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt artifacts module
~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from collections import defaultdict
from enum import Enum
from os.path import join

from .config import Config
from .utils import link_node


ArtifactType = Enum('ArtifactType', """app
                                       object
                                       binary
                                       gen_py
                                       gen_cc
                                       gen_h
                                       gen_go
                                       proto
                                       proto_descriptor
                                       custom_installer
                                       docker_image
                                       """)


class ArtifactStore:
    """A class for keeping track of artifacts produced during build.

    Generally, the class tracks only *files* produced by *builders* during
    their build func execution.

    Supported artifact types:
    - "app"     An artifact that is just a copy of a file that exists under
                the project tree (not generated / compiled).
                Such artifacts will not be cached.
                When building Docker images, such artifacts are copied under
                the path /usr/src/app within the image.
    - "object"  A compiled object file that is needed by linking-builders, but
                not by other dependent targets, and is not copied to built
                Docker images. Object files are cached as outputs.
    - "binary"  A compiled executable binary file, that will be copied under
                the path /usr/src/bin in dependent Docker images.
                Binary files are cached as outputs.
    - "gen_py"  A generated Python file (e.g. from Proto's) that is used by
                dependent test targets and Docker images. It will be copied
                under the path /usr/src/gen. Cached as outputs.
    - "gen_cc"  A generated C/C++ source file (e.g. from Proto's) that is used
                only by directly dependent C++ builders (used to extend
                sources). Cached as outputs.
    - "gen_h"   A generated C/C++ header file (e.g. from Proto's) that is used
                by any dependent C++ builder (direct or indirect) to extend
                the target headers. Cached as ouputs.
    - "gen_go"  A generated Go source file (e.g. from Proto's) that is used
                only by directly and indirect dependents Go builders (used to
                extend sources). Cached as outputs.


    An artifact is added as a specific type, with a source path (pointing to
    the location of the artifact file relative to project root), and an
    optional destination path (if different from source path).
    """

    # Mapping from artifact type to sub-dir under /usr/src
    type_to_dir = {
        ArtifactType.app: 'app',
        ArtifactType.binary: 'bin',
        ArtifactType.gen_py: 'gen',
        ArtifactType.gen_cc: '',
        ArtifactType.gen_h: '',
        ArtifactType.gen_go: 'go',
        ArtifactType.proto_descriptor: 'gen'
    }

    def __init__(self):
        self._artifacts = defaultdict(dict)

    def add(self, artifact_type: ArtifactType, src_path: str,
            dst_path: str=None):
        """Add an artifact of type `artifact_type` at `src_path`.

        `src_path` should be the path of the file relative to project root.
        `dst_path`, if given, is the desired path of the artifact in dependent
        targets, relative to its base path (by type).
        """
        if dst_path is None:
            dst_path = src_path
        other_src_path = self._artifacts[artifact_type].setdefault(
            dst_path, src_path)
        if src_path != other_src_path:
            raise RuntimeError(
                '{} artifact with dest path {} exists with different src '
                'path: {} != {}'.format(artifact_type, dst_path, src_path,
                                        other_src_path))

    def extend(self, artifact_type: ArtifactType, src_paths: list):
        """Add all `src_paths` as artifact of type `artifact_type`."""
        for src_path in src_paths:
            self.add(artifact_type, src_path, src_path)

    def get(self, artifact_type: ArtifactType) -> dict:
        """Return artifacts dict of type `artifact_type`."""
        return self._artifacts[artifact_type]

    def get_all(self) -> dict:
        return self._artifacts

    def reset(self):
        """Clear internal artifacts store."""
        self._artifacts.clear()

    def link_types(self, base_dir: str, types: list, conf: Config) -> int:
        """Link all artifacts with types `types` under `base_dir` and return
           the number of linked artifacts."""
        num_linked = 0
        for kind in types:
            artifact_map = self._artifacts.get(kind)
            if not artifact_map:
                continue
            num_linked += self._link(join(base_dir, self.type_to_dir[kind]),
                                     artifact_map, conf)
        return num_linked

    def link_for_image(self, base_dir: str, conf: Config) -> int:
        """Link all artifacts required for a Docker image under `base_dir` and
           return the number of linked artifacts."""
        return self.link_types(
            base_dir,
            [ArtifactType.app, ArtifactType.binary, ArtifactType.gen_py],
            conf)

    def _link(self, base_dir: str, artifact_map: dict, conf: Config):
        """Link all artifacts in `artifact_map` under `base_dir` and return
           the number of artifacts linked."""
        num_linked = 0
        for dst, src in artifact_map.items():
            abs_src = join(conf.project_root, src)
            abs_dest = join(conf.project_root, base_dir, dst)
            link_node(abs_src, abs_dest)
            num_linked += 1
        return num_linked



# -*- coding: utf-8 -*-

# Copyright 2020 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt Go Builder
~~~~~~~~~~~~~~~

:author: Itamar Ostricher

TODO: libs, external libs
TODO: does this even work with non-flat source file tree??
"""
import shutil
from ..config import YSETTINGS_FILE
from os import listdir, remove
from os.path import isfile, join, relpath, dirname, basename

from yabt.docker import extend_runtime_params, format_docker_run_params
from ..artifact import ArtifactType as AT
from .dockerapp import build_app_docker_and_bin, register_app_builder_sig
from ..extend import (
    PropType as PT, register_build_func, register_builder_sig,
    register_test_func, register_manipulate_target_hook)
from ..logging import make_logger
from ..target_utils import split
from ..utils import link_files, link_node, rmtree, yprint

logger = make_logger(__name__)


register_app_builder_sig('GoApp', [('main', PT.Target)])


@register_manipulate_target_hook('GoApp')
def go_app_manipulate_target(build_context, target):
    logger.debug('Injecting {} to deps of {}',
                 target.props.base_image, target.name)
    target.deps.append(target.props.base_image)
    if target.props.main and target.props.main not in target.deps:
        logger.debug('Injecting {} to deps of {}',
                     target.props.main, target.name)
        target.deps.append(target.props.main)


@register_build_func('GoApp')
def go_app_builder(build_context, target):
    """Pack a Go binary as a Docker image with its runtime dependencies."""
    yprint(build_context.conf, 'Build GoApp', target)
    prog = build_context.targets[target.props.main]
    binary = list(prog.artifacts.get(AT.binary).keys())[0]
    entrypoint = ['/usr/src/bin/' + binary]
    build_app_docker_and_bin(
        build_context, target, entrypoint=entrypoint)


# Common Go builder signature terms
GO_COMMON_SIG = [
    ('sources', PT.FileList),
    ('in_buildenv', PT.Target),
    ('cmd_env', None),
    ('build_flags', PT.StrList, None),
]
GO_BIN_SIG = GO_COMMON_SIG + [
    ('mod_file', PT.File, None),
]
GO_TEST_SIG = GO_BIN_SIG + [
    ('test_flags', PT.StrList, None),
]

register_builder_sig('GoProg', GO_BIN_SIG)


@register_manipulate_target_hook('GoProg')
def go_prog_manipulate_target(build_context, target):
    target.buildenv = target.props.in_buildenv


@register_build_func('GoProg')
def go_prog_builder(build_context, target):
    """Build a Go binary executable"""
    go_builder_internal(build_context, target, command='build')


register_builder_sig('GoPackage', GO_COMMON_SIG)


@register_manipulate_target_hook('GoPackage')
def go_package_manipulate_target(build_context, target):
    target.buildenv = target.props.in_buildenv


@register_build_func('GoPackage')
def go_package_builder(build_context, target):
    """Build a Go package"""
    go_builder_internal(build_context, target, command='build',
                        is_binary=False)


register_builder_sig('GoTest', GO_TEST_SIG)


@register_manipulate_target_hook('GoTest')
def go_test_manipulate_target(build_context, target):
    target.tags.add('testable')
    target.buildenv = target.props.in_buildenv


@register_build_func('GoTest')
def go_test_builder(build_context, target):
    """Test a Go test"""
    go_builder_internal(build_context, target, command='test')


@register_test_func('GoTest')
def go_test_tester(build_context, target):
    """Run a Go test executable"""
    yprint(build_context.conf, "Run GoTest", target)
    workspace_dir = build_context.get_workspace('GoTest', target.name)
    buildenv_workspace = build_context.conf.host_to_buildenv_path(
        workspace_dir)
    test_cmd = [join(buildenv_workspace, *split(target.name))]
    test_cmd.append('--test.v')
    test_cmd.extend(target.props.test_flags)
    run_params = extend_runtime_params(
        target.props.runtime_params,
        build_context.walk_target_deps_topological_order(target),
        build_context.conf.runtime_params, True)
    build_context.run_in_buildenv(
        target.props.in_buildenv, test_cmd,
        run_params=format_docker_run_params(run_params))


def rm_all_but_go_mod(workspace_dir):
    for fname in listdir(workspace_dir):
        if fname == 'go.mod':
            continue
        filepath = join(workspace_dir, fname)
        if isfile(filepath):
            remove(filepath)
        else:
            rmtree(filepath)


def generate_user_go_mod(build_context, target, build_cmd_env, go_module,
                         user_mod_path):
    user_mod_dir = dirname(user_mod_path)
    user_mod_file = basename(user_mod_path)
    build_context.run_in_buildenv(
        target.props.in_buildenv,
        ['go', 'mod', 'init', go_module],
        build_cmd_env,
        work_dir=build_context.conf.host_to_buildenv_path(user_mod_dir)
    )
    if user_mod_file != 'go.mod':
        shutil.move(join(user_mod_dir, 'go.mod'), user_mod_path)


def go_builder_internal(build_context, target, command, is_binary=True):
    """
    Build or test a Go package or Go binary executable.

    :param is_binary: True if binary artifact.
    :param build_context:
    :param target:
    :param command: Can be either 'build' or 'test'
    :return: Nothing

    Build of Go targets is done building a module tree under the target
    workspace.
    We link all Go source files from all dependencies to the module tree.
    We link all Go proto generated files to 'proto' sub-module under the
    module tree.
    We create a go.mod file in module root with a global module-name.
    We set the first dir in GOPATH to be yabtwork/go so that all downloaded
    packages are managed in the user machine and not inside the ephemeral
    docker.

    TODOs:
      - Support multiple modules in the same repo. Currently only one module
        is supported. Better than to define go_module in YSettings and not in
        the different targets. Using different go_modules is untested, and
        will probably will not be able to import packages from one module in
        another one.

    """

    builder_name = target.builder_name
    yprint(build_context.conf, command, builder_name, target)

    go_module = build_context.conf.get('go_module', None)
    if not go_module:
        raise KeyError("Must specify go_module in {} common_conf".format(
            YSETTINGS_FILE
        ))

    workspace_dir = build_context.get_workspace(builder_name, target.name)
    rmtree(workspace_dir)

    buildenv_workspace = build_context.conf.host_to_buildenv_path(
        workspace_dir
    )
    target_sources = []
    for src in target.props.sources:
        workspace_src = join("src", src)
        target_file = join(workspace_dir, workspace_src)
        link_node(join(build_context.conf.project_root, src),
                  target_file)
        target_sources.append(workspace_src)
    buildenv_sources = [
        join(buildenv_workspace, src) for src in target_sources
    ]

    files_to_link = []
    for dep in build_context.generate_all_deps(target):
        files_to_link.extend(filter(lambda x: x.endswith('.go'),
                                    dep.props.get('sources', [])))
        artifact_map = dep.artifacts.get(AT.gen_go)
        if not artifact_map:
            continue
        for dst, src in artifact_map.items():
            target_file = join(workspace_dir, dst)
            link_node(join(build_context.conf.project_root, src),
                      target_file)

    if len(files_to_link) > 0:
        link_files(files_to_link, workspace_dir, None, build_context.conf)

    download_cache_dir = build_context.conf.host_to_buildenv_path(
        build_context.conf.get_go_packages_path())

    gopaths = [download_cache_dir]
    user_gopath = (target.props.cmd_env or {}).get('GOPATH')
    # if user didn't provide GOPATH we assumes it is /go
    # TODO(eyal): A more correct behavior will be to check the GOPATH var
    # inside the docker then to assumes it is /go.
    # This code provides a way for the user to tell us what is the correct
    # GOPATH but if it come handy we should implement looking into the docker
    gopaths.append(user_gopath if user_gopath else '/go')
    build_cmd_env = {
        'XDG_CACHE_HOME': download_cache_dir,
    }
    build_cmd_env.update(target.props.cmd_env or {})
    build_cmd_env['GOPATH'] = ':'.join(gopaths)

    go_mod_path = join(workspace_dir, 'go.mod')
    user_mod_path = \
        target.props.get('mod_file', None) or \
        build_context.conf.get('go_mod_file', None)
    if user_mod_path:
        user_mod_path = join(build_context.conf.project_root, user_mod_path)
        if not isfile(user_mod_path):
            generate_user_go_mod(build_context, target, build_cmd_env,
                                 go_module, user_mod_path)
        link_node(user_mod_path, go_mod_path)
    if not isfile(go_mod_path):
        build_context.run_in_buildenv(
            target.props.in_buildenv,
            ['go', 'mod', 'init', go_module],
            build_cmd_env,
            work_dir=buildenv_workspace
        )

    if len(buildenv_sources) > 0:
        binary = join(*split(target.name)) if is_binary else None
        binary_args = []
        if binary:
            if command == 'test':
                binary_args.append('-c')
            bin_file = join(buildenv_workspace, binary)
            binary_args.extend(['-o', bin_file])

        build_cmd = ['go', command] + target.props.build_flags + \
            binary_args + buildenv_sources
        run_params = extend_runtime_params(
            target.props.runtime_params,
            build_context.walk_target_deps_topological_order(target),
            build_context.conf.runtime_params,
            True
        )
        build_context.run_in_buildenv(target.props.in_buildenv, build_cmd,
                                      build_cmd_env,
                                      run_params=format_docker_run_params(
                                          run_params
                                      ),
                                      work_dir=buildenv_workspace)
        if binary:
            target.artifacts.add(
                AT.binary, relpath(join(workspace_dir, binary),
                                   build_context.conf.project_root), binary
            )
    else:
        logger.warn("target {} has no sources to build".format(target.name))



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt NodeJS Builders
~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from ..extend import (
    PropType as PT, register_build_func, register_builder_sig,
    register_manipulate_target_hook)
from ..utils import yprint


register_builder_sig(
    'NpmPackage',
    [('package', PT.str),
     ('version', PT.str, None),
     ('global_install', PT.bool, True),
     ])


def format_npm_specifier(target):
    if target.props.version:
        return '{0.package}@{0.version}'.format(target.props)
    return '{0.package}'.format(target.props)


@register_build_func('NpmPackage')
def npm_package_builder(build_context, target):
    yprint(build_context.conf, 'Fetch and cache NPM package', target)


@register_manipulate_target_hook('NpmPackage')
def npm_package_manipulate_target(build_context, target):
    target.tags.add('npm-installable')



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt config
~~~~~~~~~~~

:author: Itamar Ostricher
"""


import os
from os.path import abspath, relpath
from pathlib import Path

from ostrich.utils.collections import listify
from ostrich.utils.text import get_safe_path

from .extend import Plugin
from .logging import configure_logging
from .scm import ScmManager
from .utils import norm_proj_path, search_for_parent_dir


BUILD_PROJ_FILE = 'YRoot'
YCONFIG_FILE = 'YConfig'
YSETTINGS_FILE = 'YSettings'


class Config:
    """Runtime Config info class"""

    attrs_from_args = frozenset((
        'artifacts_metadata_file',
        'continue_after_fail',
        'bin_output_dir',
        'build_base_images',
        'build_file_name',
        'builders_workspace_dir',
        'cmd',
        'default_target_name',
        'docker_volume',
        'flavor',
        'force_pull',
        'jobs',
        'no_build_cache',
        'no_docker_cache',
        'no_test_cache',
        'non_interactive',
        'offline',
        'output_dot_file',
        'push',
        'runtime_params',
        'targets',
        'test_attempts',
        'verbose',
        'with_tini_entrypoint',
        # logging-related
        'loglevel', 'logtostderr', 'logtostdout',
        'show_buildenv_deps',
        'download_from_global_cache',
        'upload_to_global_cache',
        'download_tests_from_global_cache',
        'upload_tests_to_global_cache',
    ))

    def __init__(self, args, project_root_dir: str, work_dir: str,
                 settings_module=None):
        """
        :param project_root_dir: Absolute path to build project root directory.
        :param work_dir: Absolute path to working directory within project.
        """
        for slot in self.attrs_from_args:
            setattr(self, slot, getattr(args, slot))
        self.docker_pull_cmd = args.docker_pull_cmd.split()
        self.docker_push_cmd = args.docker_push_cmd.split()
        configure_logging(self)
        self.project_root = project_root_dir
        self.work_dir = work_dir
        self.scm_provider = str(args.scm_provider).lower()
        self.scm = ScmManager.get_provider(self.scm_provider, self)
        Plugin.load_plugins(self)
        self.settings = settings_module
        self.common_conf = {}
        self.flavor_conf = {}
        self.flavor_dir = get_safe_path('flavor__all__' if self.flavor is None
                                        else '{}_flavor'.format(self.flavor))
        self.policies = []

    def in_yabt_project(self) -> bool:
        return self.project_root is not None

    def get_rel_work_dir(self) -> str:
        return os.path.relpath(self.work_dir, self.project_root)

    def get_project_build_file(self) -> str:
        return os.path.join(self.project_root, BUILD_PROJ_FILE)

    def get_build_file_path(self, build_module) -> str:
        """Return a full path to the build file of `build_module`.

        The returned path will always be OS-native, regardless of the format
        of project_root (native) and build_module (with '/').
        """
        project_root = Path(self.project_root)
        build_module = norm_proj_path(build_module, '')
        return str(project_root / build_module /
                   (BUILD_PROJ_FILE if '' == build_module
                    else self.build_file_name))

    def get_go_packages_path(self) -> str:
        return os.path.join(
            self.project_root, self.builders_workspace_dir, 'go')

    def get_workspace_path(self) -> str:
        return os.path.join(
            self.project_root, self.builders_workspace_dir, self.flavor_dir)

    def get_bin_path(self) -> str:
        return os.path.join(self.project_root, self.bin_output_dir)

    def get_cache_dir(self, target, build_context) -> str:
        # TODO: consider using host-level shared cache, so it can be shared
        # between different copies of a project (multiple clones / users)
        # consider dockerized execution...
        return os.path.join(self.project_root, self.builders_workspace_dir,
                            '.cache', 'targets', target.hash(build_context))

    def get_artifacts_cache_dir(self) -> str:
        # TODO: consider using host-level shared cache, so it can be shared
        # between different copies of a project (multiple clones / users)
        # consider dockerized execution...
        return os.path.join(self.project_root, self.builders_workspace_dir,
                            '.cache', 'artifacts')

    def host_to_buildenv_path(self, host_path: str) -> str:
        # TODO: windows-containers?
        return '/'.join([
            '/project',
            Path(relpath(abspath(host_path), self.project_root)).as_posix()])

    def get(self, param: str, fallback: str) -> str:
        common_val, flavor_val = None, None
        if self.common_conf:
            common_val = self.common_conf.get(param)
        if self.flavor_conf:
            flavor_val = self.flavor_conf.get(param)
        if common_val is None and flavor_val is None:
            return fallback
        if flavor_val is not None:
            if isinstance(flavor_val, list):
                val = []
                for el in flavor_val:
                    if el == '$*':
                        val.extend(listify(common_val))
                    else:
                        val.append(el)
                return val
            return flavor_val
        return common_val or fallback



# -*- coding: utf-8 -*-

# Copyright 2018 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

:author: Itamar Ostricher
"""

import random
import pytest

from ..buildcontext import BuildContext
from ..graph import populate_targets_graph


@pytest.mark.slow
@pytest.mark.usefixtures('in_tests_project')
def test_python_tester_success(basic_conf):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['hello_pytest:greet-test']
    populate_targets_graph(build_context, basic_conf)
    build_context.build_graph(run_tests=True)


@pytest.mark.slow
@pytest.mark.usefixtures('in_tests_project')
def test_python_tester_fail(basic_conf):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['hello_pytest:greet-failing-test']
    populate_targets_graph(build_context, basic_conf)
    with pytest.raises(SystemExit):
        build_context.build_graph(run_tests=True)


@pytest.mark.slow
@pytest.mark.usefixtures('in_tests_project')
def test_python_tester_fail_no_exit(basic_conf, capsys):
    basic_conf.continue_after_fail = True
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['hello_pytest:greet-failing-test']
    populate_targets_graph(build_context, basic_conf)
    with pytest.raises(SystemExit):
        build_context.build_graph(run_tests=True)
    out, err = capsys.readouterr()
    err.encode()
    expected_error = "\x1b[31mFatal: Finished building target graph with \
fails: \n['hello_pytest:greet-failing-test']\nwhich caused the \
following to skip: \n[]\x1b[0m\n"
    expected_error.encode()
    assert err == expected_error


@pytest.mark.slow
@pytest.mark.usefixtures('in_tests_project')
def test_python_tester_fail_with_retry(basic_conf):
    build_context = BuildContext(basic_conf)
    target_name = 'hello_pytest:greet-failing-test'
    basic_conf.targets = [target_name]
    populate_targets_graph(build_context, basic_conf)
    target = build_context.targets[target_name]
    target.props.attempts = 5
    with pytest.raises(SystemExit):
        build_context.build_graph(run_tests=True)
    assert target.info['fail_count'] == 5


@pytest.mark.slow
@pytest.mark.usefixtures('in_tests_project')
def test_python_tester_flaky(basic_conf):
    build_context = BuildContext(basic_conf)
    target_name = 'hello_pytest:flaky-test'
    basic_conf.targets = [target_name]
    populate_targets_graph(build_context, basic_conf)
    target = build_context.targets[target_name]
    target.props.test_env['RANDOM_FILE'] = str(random.randint(0, 20000))
    build_context.build_graph(run_tests=True)
    assert target.info['fail_count'] == 1


@pytest.mark.slow
@pytest.mark.usefixtures('in_tests_project')
def test_python_tester_aba(basic_conf):
    build_context = BuildContext(basic_conf)
    target_a_name = 'hello_pytest:test-a'
    target_b_name = 'hello_pytest:test-b'
    basic_conf.targets = [target_a_name, target_b_name]
    populate_targets_graph(build_context, basic_conf)
    target_a = build_context.targets[target_a_name]
    target_b = build_context.targets[target_b_name]
    random_file = str(random.randint(0, 20000))
    target_a.props.test_env['RANDOM_FILE'] = random_file
    target_b.props.test_env['RANDOM_FILE'] = random_file
    build_context.build_graph(run_tests=True)
    assert (
        target_a.info['fail_count'] == 1 or target_b.info['fail_count'] == 1)
    assert (
        target_a.info['fail_count'] == 0 or target_b.info['fail_count'] == 0)



# -*- coding: utf-8 -*-

# Copyright 2018 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt Policy module
~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""

import json
from os import makedirs
from os.path import isdir, isfile, join, relpath, split
import shutil
from time import time

from ostrich.utils.text import get_safe_path

from .artifact import ArtifactType as AT
from .config import Config
from .docker import get_image_name, handle_build_cache, tag_docker_image
from .graph import get_descendants
from .logging import make_logger
from .target_utils import ImageCachingBehavior, Target
from .utils import hash_tree, rmnode, rmtree


logger = make_logger(__name__)


# source: https://opensource.org/licenses/alphabetical
KNOWN_LICENSES = frozenset((
    '0BSD',
    'AAL',
    'AFL-3.0',
    'AGPL-3.0',
    'APL-1.0',
    'APSL-2.0',
    'Apache-2.0',
    'Artistic-2.0',
    'BSD-2-Clause',
    'BSD-2-Clause-Patent',
    'BSD-3-Clause',
    'BSL-1.0',
    'CATOSL-1.1',
    'CDDL-1.0',
    'CECILL-2.1',
    'CNRI portion of Python License',
    'CNRI-Python',
    'CPAL-1.0',
    'CUA-OPL-1.0',
    'ECL-2.0',
    'EFL-2.0',
    'EPL-1.0',
    'EPL-2.0',
    'EUDatagrid',
    'EUPL-1.1',
    'Entessa',
    'Fair',
    'Frameworx-1.0',
    'GPL-2.0',
    'GPL-3.0',
    'HPND',
    'IPA',
    'IPL-1.0',
    'ISC',
    'LGPL-2.1',
    'LGPL-3.0',
    'LPL-1.02',
    'LPPL-1.3c',
    'LiLiQ-P',
    'LiLiQ-R',
    'LiLiQ-R+',
    'MIT',
    'MPL-1.0',
    'MPL-1.1',
    'MPL-2.0',
    'MS-PL',
    'MS-RL',
    'MirOS',
    'Motosoto',
    'Multics',
    'NASA-1.3',
    'NCSA',
    'NGPL',
    'NPOSL-3.0',
    'NTP',
    'Naumen',
    'Nokia',
    'OCLC-2.0',
    'OFL-1.1',
    'OGTSL',
    'OSL-3.0',
    'PHP-3.0',
    'PostgreSQL',
    'Python-2.0',
    'QPL-1.0',
    'RPL-1.5',
    'RPSL-1.0',
    'RSCPL',
    'SPL-1.0',
    'SimPL-2.0',
    'Sleepycat',
    'UPL',
    'VSL-1.0',
    'W3C',
    'WXwindows',
    'Watcom-1.0',
    'Xnet',
    'ZPL-2.0',
    'Zlib',
    # *** manual additions *** #
    'Other',
    'Commercial',
))


def standard_licenses_only(build_context, target) -> str:
    """A policy function for allowing specifying only known licenses.

    Return error message (string) if policy for `target` is violated,
    otherwise return `None`.

    To apply in project, include this function in the ilst returned by the
    `get_policies` function implemented in the project `YSettings` file.

    See example in tests/errors.
    """
    for license_name in target.props.license:
        if license_name not in KNOWN_LICENSES:
            # TODO: include suggestion for similar known license
            return 'Unknown license: {}'.format(license_name)
    return None


def whitelist_licenses_policy(policy_name: str, allowed_licenses: set):
    """A policy factory for making license-based whitelist policies.

    To apply in project, include the function returned from this factory
    in the ilst returned by the `get_policies` function implemented in the
    project `YSettings` file.

    The factory returns a policy function named
    `whitelist_{policy_name}_licenses` that applies to targets with
    `policy_name` in their policies list.
    The returned policy asserts that all licenses contained in the target
    (including through explicit & implicit dependencies) are in the whitelist
    defined by `allowed_licenses`.

    See example in tests/errors.
    """

    def policy_func(build_context, target):
        """whitelist_{policy_name}_licenses policy function.

        Return error message (string) if policy for `target` is violated,
        otherwise return `None`.
        """
        if policy_name in target.props.policies:
            licenses = set(target.props.license)
            for dep in build_context.generate_all_deps(target):
                licenses.update(dep.props.license)
            licenses.difference_update(allowed_licenses)
            if licenses:
                return 'Invalid licenses for {} policy: {}'.format(
                    policy_name, ', '.join(sorted(licenses)))
        return None

    policy_func.__name__ = 'whitelist_{}_licenses'.format(policy_name)
    return policy_func



# -*- coding: utf-8 -*-

# Copyright 2019 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt caching random tests
~~~~~~~~~~~~~~~~~~

:author: Dana Shamir
"""

import json
import networkx as nx
import os
from os.path import join, dirname, abspath, getmtime, isfile
import pytest
import random
import shutil
import string
from unittest.mock import Mock

from conftest import reset_parser
from yabt.builders import cpp
from yabt.extend import Plugin
from . import config, cli, extend
from .buildcontext import BuildContext
from .global_cache import SUMMARY_FILE, ARTIFACTS_FILE, TARGETS_DIR, \
    ARTIFACTS_DIR
from .graph import populate_targets_graph
from .logging import make_logger
from .test_utils import generate_random_dag

NUM_TARGETS = 10
NUM_TESTS = 20

TMPL_DIR = join(dirname(abspath(__file__)), '..', 'tests', 'data',
                'caching')
CPP_TMPL = 'cpp_prog.cc.tmpl'
CPP_TEST_TMPL = 'cpp_test.cc.tmpl'
FAILING_TEST_TMPL = 'cpp_failing_test.cc.tmpl'
CPP_TARGET = """CppProg(
    '{}',
    sources='{}',
    in_buildenv=':builder', deps={}
)"""
CPP_TEST_TARGET = """CppGTest(
    '{}',
    sources='{}',
    in_buildenv=':builder-with-gtest',
    deps={}
)"""
YROOT_TMPL = 'YRoot.tmpl'
INSTALL_GTEST_SCRIPT = 'install-gtest.sh'
INSTALL_LLVM_REPO_KEY_SCRIPT = 'install-llvm-repo-key.sh'
YSETTINGS = 'YSettings'

TARGET_TYPES = {
    'CppProg': (CPP_TMPL, CPP_TARGET),
    'CppGTest': (CPP_TEST_TMPL, CPP_TEST_TARGET)
}

GLOBAL_CACHE_DIR = '.global_cache'

logger = make_logger(__name__)


class ProjectContext:
    def __init__(self):
        self.targets = {}
        self.test_targets = []
        self.targets_graph = None
        self.last_modified = {}
        self.last_run_tests = {}
        self.conf = None


def generate_random_project(size) -> ProjectContext:
    project = ProjectContext()
    targets = [random_string() for _ in range(size)]
    project.targets_graph = generate_random_dag(targets)
    for target in targets:
        target_type = random.choice(list(TARGET_TYPES.keys()))
        generate_file(target, target_type)
        project.targets[target] = target_type
        if target_type == 'CppGTest':
            project.test_targets.append(target)
    generate_yroot(project)
    with open(join(TMPL_DIR, YSETTINGS + '.tmpl'), 'r') as tmpl:
        ysettings = tmpl.read().format(GLOBAL_CACHE_DIR)
    with open(YSETTINGS, 'w') as ysettings_file:
        ysettings_file.write(ysettings)
    return project


def random_string():
    return ''.join([random.choice(string.ascii_letters + string.digits)
                    for _ in range(random.randint(20, 40))])


def generate_yroot(project: ProjectContext):
    yroot = []
    for target, target_type in project.targets.items():
        yroot.append(TARGET_TYPES[target_type][1].format(
            target, get_file_name(target),
            get_dependencies(target, project.targets_graph)))
    with open(join(TMPL_DIR, YROOT_TMPL), 'r') as yroot_tmpl_file:
        yroot_data = yroot_tmpl_file.read()
    with open(config.BUILD_PROJ_FILE, 'w') as yroot_file:
        yroot_file.write(yroot_data + '\n\n'.join(yroot))
    shutil.copyfile(join(TMPL_DIR, INSTALL_GTEST_SCRIPT), INSTALL_GTEST_SCRIPT)
    shutil.copyfile(join(TMPL_DIR, INSTALL_LLVM_REPO_KEY_SCRIPT),
                    INSTALL_LLVM_REPO_KEY_SCRIPT)


def get_dependencies(target, target_graph):
    return [':' + dep for dep, other_target in target_graph.edges
            if other_target == target]


def generate_file(target, target_type, string_to_print=None):
    if string_to_print is None:
        string_to_print = target
    with open(join(TMPL_DIR, TARGET_TYPES[target_type][0]), 'r') as tmpl:
        code = tmpl.read().format(string_to_print)
    with open(get_file_name(target), 'w') as target_file:
        target_file.write(code)


def get_file_name(target):
    return join(target + '.cc')


def init_project(project):
    reset_parser()
    project.conf = cli.init_and_get_conf(['--non-interactive',
                                          '--continue-after-fail', 'build',
                                          '--upload-to-global-cache',
                                          '--download-from-global-cache',
                                          '--download-tests-from-global-cache',
                                          '--upload-tests-to-global-cache'])
    extend.Plugin.load_plugins(project.conf)
    project.conf.targets = [':' + target for target in project.targets.keys()]
    build_context = BuildContext(project.conf)
    populate_targets_graph(build_context, project.conf)
    return build_context


def rebuild(project: ProjectContext):
    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    check_modified_targets(project, build_context, [])


def rebuild_after_modify(project: ProjectContext):
    target_to_change = random.choice(list(project.targets.keys()))
    logger.info('modifing target: {}'.format(target_to_change))
    generate_file(target_to_change, project.targets[target_to_change],
                  random_string())
    build_context = init_project(project)
    build_context.build_graph(run_tests=True)

    targets_to_build = nx.descendants(project.targets_graph, target_to_change)
    targets_to_build.add(target_to_change)
    check_modified_targets(project, build_context, targets_to_build)


def delete_file_and_return_no_modify(project: ProjectContext):
    target_to_delete = random.choice(list(project.targets.keys()))
    logger.info('deleting and returning the same for target: {}'
                .format(target_to_delete))
    file_name = get_file_name(target_to_delete)
    with open(file_name) as target_file:
        curr_content = target_file.read()
    os.remove(file_name)
    with open(file_name, 'w') as target_file:
        target_file.write(curr_content)

    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    check_modified_targets(project, build_context, [])


def add_dependency(project: ProjectContext):
    new_target = random_string()
    target_type = random.choice(list(TARGET_TYPES.keys()))
    logger.info('adding target: {} of type: {}'.format(new_target,
                                                       target_type))
    project.targets[new_target] = target_type
    project.last_modified[new_target] = None
    if target_type == 'CppGTest':
        project.test_targets.append(new_target)
        project.last_run_tests[new_target] = None
    project.conf.targets.append(':' + new_target)
    generate_file(new_target, target_type)

    project.targets_graph.add_node(new_target)
    targets = list(project.targets.keys())
    project.targets_graph.add_edges_from(
        (new_target, targets[i]) for i in range(len(targets) - 1)
        if random.random() > 0.8 and targets[i] != new_target)
    generate_yroot(project)

    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    targets_to_build = nx.descendants(project.targets_graph, new_target)
    targets_to_build.add(new_target)
    check_modified_targets(project, build_context, targets_to_build)


def download_from_global_cache(project: ProjectContext):
    target = random.choice(list(project.targets.keys()))
    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    check_modified_targets(project, build_context, [])
    cache_dir = project.conf.get_cache_dir(
        build_context.targets[':' + target], build_context)
    logger.info('removing cache from: {} of target: {}'.format(cache_dir,
                                                               target))
    shutil.rmtree(cache_dir)

    cppgtest_mock = Mock()
    old_cppgtest_builder = Plugin.builders['CppGTest']
    old_cppgtest_hook = Plugin.hooks['manipulate_target']['CppGTest']
    Plugin.remove_builder('CppGTest')
    Plugin.builders['CppGTest'] = cppgtest_mock
    cppgtest_mock.func.side_effect = cpp.cpp_gtest_builder
    Plugin.hooks['manipulate_target']['CppGTest'] = old_cppgtest_hook

    build_context.build_graph(run_tests=True)
    cppgtest_mock.test_func.assert_not_called()
    check_modified_targets(project, build_context, [], [target])

    Plugin.remove_builder('CppGTest')
    Plugin.builders['CppGTest'] = old_cppgtest_builder
    Plugin.hooks['manipulate_target']['CppGTest'] = old_cppgtest_hook


def no_cache_at_all(project: ProjectContext):
    target_name = random.choice(list(project.targets.keys()))
    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    check_modified_targets(project, build_context, [])
    target = build_context.targets[':' + target_name]
    logger.info('removing local and global cache of target: {}'.format(
        target_name))
    shutil.rmtree(project.conf.get_cache_dir(target, build_context))
    shutil.rmtree(join(GLOBAL_CACHE_DIR, TARGETS_DIR,
                       target.hash(build_context)))
    build_context.build_graph(run_tests=True)
    targets_to_build = nx.descendants(project.targets_graph, target_name)
    targets_to_build.add(target_name)
    check_modified_targets(project, build_context, targets_to_build)


def failing_test(project: ProjectContext):
    test_to_fail = random.choice(project.test_targets)
    logger.info('Making target: {} fail'.format(test_to_fail))
    with open(join(TMPL_DIR, FAILING_TEST_TMPL), 'r') as tmpl:
        code = tmpl.read().format(test_to_fail)
    with open(get_file_name(test_to_fail), 'w') as target_file:
        target_file.write(code)

    build_context = init_project(project)
    with pytest.raises(SystemExit):
        build_context.build_graph(run_tests=True)
    test_cache = get_test_cache(project.conf, test_to_fail, build_context)
    assert not os.path.isfile(test_cache), \
        "test: {} is failing but was put into cache".format(test_to_fail)

    generate_file(test_to_fail, project.targets[test_to_fail], random_string())
    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    targets_to_build = nx.descendants(project.targets_graph, test_to_fail)
    targets_to_build.add(test_to_fail)
    check_modified_targets(project, build_context, targets_to_build)


def randomly_delete_global_cache(project: ProjectContext, file_to_delete):
    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    paths_to_delete, targets_to_delete = get_random_targets_to_delete(
        project, build_context)
    for path in paths_to_delete:
        if isfile(join(path, file_to_delete)):
            os.remove(join(path, file_to_delete))

    remove_targets_from_local_cache(build_context, project, targets_to_delete)
    build_and_check_built(project, targets_to_delete)
    logger.info('deleted targets: {}, paths deleted: {}'.format(
        targets_to_delete, paths_to_delete))


def remove_targets_from_local_cache(build_context, project, targets_to_delete):
    for target in targets_to_delete:
        shutil.rmtree(project.conf.get_cache_dir(
            build_context.targets[':' + target], build_context))


def randomly_delete_summary_from_global_cache(project: ProjectContext):
    randomly_delete_global_cache(project, SUMMARY_FILE)


def randomly_delete_artifacts_desc_from_global_cache(project: ProjectContext):
    randomly_delete_global_cache(project, ARTIFACTS_FILE)


def randomly_delete_artifacts_from_global_cache(project: ProjectContext):
    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    paths_to_delete, targets_to_build = get_random_artifacts_to_delete(
        project, build_context)
    for path in paths_to_delete:
        os.remove(path)
    remove_targets_from_local_cache(build_context, project, targets_to_build)

    build_and_check_built(project, targets_to_build)


def build_and_check_built(project, modified_targets):
    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    targets_to_build = set(modified_targets)
    for target in modified_targets:
        targets_to_build.update(nx.descendants(project.targets_graph, target))
    check_modified_targets(project, build_context, targets_to_build)


def get_random_targets_to_delete(project: ProjectContext,
                                 build_context: BuildContext):
    targets_dir = join(GLOBAL_CACHE_DIR, TARGETS_DIR)
    all_targets = os.listdir(targets_dir)
    paths_to_delete = [join(targets_dir, filename) for filename in
                       random.sample(all_targets, len(all_targets) // 10)]
    targets_to_delete = []
    for path in paths_to_delete:
        if isfile(join(path, SUMMARY_FILE)):
            with open(join(path, SUMMARY_FILE), 'rb') as summary_file:
                summary = json.loads(summary_file.read().decode('utf-8'))
            target = summary['name'].strip(':')
            if target not in project.targets:
                paths_to_delete.remove(path)
            elif build_context.targets[summary['name']].hash(build_context) \
                    == os.path.split(path)[-1]:
                targets_to_delete.append(target)
    return paths_to_delete, targets_to_delete


def get_random_artifacts_to_delete(project: ProjectContext,
                                   build_context: BuildContext):
    targets_dir = join(GLOBAL_CACHE_DIR, TARGETS_DIR)
    artifacts_dir = join(GLOBAL_CACHE_DIR, ARTIFACTS_DIR)
    all_artifacts = os.listdir(artifacts_dir)
    artifacts_to_delete = set(random.sample(all_artifacts,
                                            len(all_artifacts) // 10))

    targets_to_build = set()
    for path in [join(targets_dir, file) for file in os.listdir(targets_dir)]:
        if isfile(join(path, ARTIFACTS_FILE)) \
                and isfile(join(path, SUMMARY_FILE)):
            with open(join(path, ARTIFACTS_FILE), 'rb') as f:
                artifact_desc = json.loads(f.read().decode('utf-8'))
            with open(join(path, SUMMARY_FILE), 'rb') as f:
                summary = json.loads(f.read().decode('utf-8'))
            target = summary['name'].strip(':')
            for type_name, artifact_list in artifact_desc.items():
                for artifact in artifact_list:
                    if artifact['hash'] in artifacts_to_delete:
                        if target not in project.targets:
                            artifacts_to_delete.remove(artifact['hash'])
                        elif build_context.targets[summary['name']].hash(
                                build_context) == os.path.split(path)[-1]:
                            targets_to_build.add(target)

    paths_to_delete = [join(artifacts_dir, filename) for filename in
                       artifacts_to_delete]
    return paths_to_delete, targets_to_build


def check_modified_targets(project: ProjectContext, build_context,
                           targets_to_build, tests_to_build=None):
    for target, target_type in project.targets.items():
        last_modified = get_last_modified(project.conf, target, target_type)
        if target in targets_to_build:
            assert last_modified != project.last_modified[target], \
                "target: {} was supposed to be built again and" \
                " wasn't".format(target)
            project.last_modified[target] = last_modified
        else:
            assert last_modified == project.last_modified[target], \
                "target: {} was modified and it wasn't supposed" \
                " to".format(target)

    tests_to_build = tests_to_build or targets_to_build
    for target in project.test_targets:
        last_run = getmtime(get_test_cache(project.conf, target,
                                           build_context))
        if target in tests_to_build:
            assert last_run != project.last_run_tests[target],\
                "test: {} was supposed to run again but wasn't".format(target)
            project.last_run_tests[target] = last_run
        else:
            assert last_run == project.last_run_tests[target],\
                "test: {} was rerun and it wasn't supposed to".format(target)


def get_last_modified(basic_conf, target, target_type):
    return getmtime(join(basic_conf.builders_workspace_dir, 'release_flavor',
                         target_type, '_' + target, target + '.o'))


def get_test_cache(basic_conf, target, build_context):
    return join(basic_conf.get_cache_dir(build_context.targets[':' + target],
                                         build_context), 'tested.json')


@pytest.mark.slow
def test_caching(tmp_dir):
    project = generate_random_project(NUM_TARGETS)

    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    logger.info('done first build')

    for target, target_type in project.targets.items():
        project.last_modified[target] = get_last_modified(project.conf, target,
                                                          target_type)
    for target in project.test_targets:
        project.last_run_tests[target] = getmtime(get_test_cache(
            project.conf, target, build_context))

    tests = [rebuild, rebuild_after_modify, delete_file_and_return_no_modify,
             add_dependency, failing_test, download_from_global_cache,
             no_cache_at_all, randomly_delete_summary_from_global_cache,
             randomly_delete_artifacts_desc_from_global_cache,
             randomly_delete_artifacts_from_global_cache]
    for i in range(NUM_TESTS):
        test_func = random.choice(tests)
        logger.info('starting build number: {} with func: {}'.format(
            i + 2, test_func.__name__))
        test_func(project)



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt Build File utils
~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from pathlib import Path
from .config import Config


def to_build_module(build_file_path: str, conf: Config) -> str:
    """Return a normalized build module name for `build_file_path`."""
    build_file = Path(build_file_path)
    root = Path(conf.project_root)
    return build_file.resolve().relative_to(root).parent.as_posix().strip('.')
